-- TO DO:
-- Support 6 and 8 digit colors so alpha can optionally be supplied

--====================================================================================================================--
--
-- COLORIZER
-- Provides standard, predefined colors and functionality for working with colors
--
-- Colorizers can be instantiated with TweakerColorizer:new()
-- Tweaker's standard colorizer is accessible via Tweaker.Colorizer
-- Various parts of Tweaker define their own Colorizers
-- 
--====================================================================================================================--

-- Local Data ----------------------------------------------------------------------------------------------------------

local util = Tweaker.util
local string = util.string
local table = util.table

local colorizerColors = {}

--====================================================================================================================--

-- The Colorizer Object ------------------------------------------------------------------------------------------------

local Colorizer = {}

colorizerColors[Colorizer] = {
	red     = "ff0000", lightred     = "ff8080", darkred     = "800000",
	orange  = "ff8000", lightorange  = "ffc080", darkorange  = "804000",
	yellow  = "ffff00", lightyellow  = "ffff80", darkyellow  = "808000",
	green   = "00ff00", lightgreen   = "80ff80", darkgreen   = "008000",
	blue    = "0000ff", lightblue    = "8080ff", darkblue    = "000080",
	violet  = "8000ff", lightviolet  = "c080ff", darkviolet  = "400080",
	
	cyan    = "00ffff", lightcyan    = "80ffff", darkcyan    = "008080",
	magenta = "ff00ff", lightmagenta = "ff80ff", darkmagenta = "800080",
	
	gray    = "808080", lightgray    = "c0c0c0", darkgray    = "404040",
	white   = "ffffff",
	black   = "000000",
}

------------------------------
-- Properties
------------------------------

function Colorizer:GetColor(name)
	name = string.lower(name)
	local customColors = colorizerColors[self]
	
	local color = customColors and customColors[name]
	if color then return color end
	
	local parent = getmetatable(self)
	while parent do
		customColors = colorizerColors[parent]
		color = customColors and customColors[name]
		if color then return color end
	
		parent = getmetatable(parent)
	end
	
	return nil
end

function Colorizer:GetColors(...)
	local n = select('#', ...)
	
	local colors
	
	if n == 0 then
		
		colors = self:GetCustomColors() or {}
		
		local parent = getmetatable(self)
		while parent do
			local customs = colorizerColors[parent]
			if customs then
				for name,color in pairs(customs) do
					if not colors[name] then colors[name] = color end
				end
			end
		
			parent = getmetatable(parent)
		end
		
		return colors
		
	elseif n == 1 and type(select(1, ...)) == "table" then
		
		colors = {}
		local colorsToGet = select(1, ...)
		
		for i = 1,table.getn(colorsToGet) do
			local name = colorsToGet[i]
			colors[name] = self:GetColor(name)
		end
		
		return colors
		
	else
		
		colors = {}
		
		for i = 1,n do
			local name = string.lower(select(i,...))
			table.insert(colors, self:GetColor(name))
		end
		
		return unpack(colors)
		
	end
end

function Colorizer:GetCustomColors()
	local customColors = colorizerColors[self]
	if not customColors then return end
	
	local colors = {}
	for name,color in pairs(customColors) do
		colors[name] = color
	end
	
	return colors
end

function Colorizer:SetColor(name, color)
	name = string.lower(name)
	colorizerColors[self] = colorizerColors[self] or {}
	colorizerColors[self][name] = color
end

function Colorizer:SetColors(colors)
	for name,color in pairs(colors) do
		self:SetColor(name, color)
	end
end

------------------------------
-- Utility Functions
------------------------------

function Colorizer:ToHexString(r,g,b)
	return string.format("%02x%02x%02x", r,g,b)
end

function Colorizer:ToDecimal(hex)
	return tonumber("0x"..hex)
end

------------------------------
-- Functions
------------------------------

function Colorizer:Apply(color, str, ...)
	if not str then return end
	
	if (...) then str = string.format(str, ...) end
	
	color = self:GetColor(color)
	if color then str = "|cff"..color..str.."|r" end
	
	return str
end

function Colorizer:RegexApply(color, regex, str, ...)
	if not str then return end
	
	if (...) then str = string.format(str) end
	
	color = self:GetColor(color)
	if color then
		local builtstr = ""
		local laststop = 1
		local start,stop = string.find(str, regex)
		while start do
			local pre = string.sub(str, laststop+1, start-1)
			local mid = string.sub(str, start, stop)
			builtstr = builtstr..pre.."|cff"..color..mid.."|r"
			laststop = stop
			start,stop = string.find(str, regex, stop+1)
		end
		builtstr = builtstr..string.sub(str, laststop+1)
		str = builtstr
		
	end
	
	return str
end

------------------------------
-- Constructor
------------------------------

function Colorizer:new(colors)
	local colorizer = {}
	
	setmetatable(colorizer, self)
	self.__index = self
	
	if colors then colorizer:SetColors(colors) end
	
	return colorizer
end

--====================================================================================================================--

-- Global --------------------------------------------------------------------------------------------------------------

TweakerColorizer = Colorizer

Tweaker.Colorizer = Colorizer:new()

--====================================================================================================================--
