--====================================================================================================================--
--
-- LOGGER
-- Some concepts are adapted from the log4j java module
--
-- Loggers can be instantiated with TweakerLogger:new()
-- Tweaker's logger is accessible via Tweaker.log
-- 
--====================================================================================================================--

-- Local Data ----------------------------------------------------------------------------------------------------------

local util = Tweaker.util
local string = util.string
local table = util.table

-- Levels of logging are adapted from the log4j java module
local logLevelMatrix = table.dupekv({
	["all"]   = 1,  -- All logging, identical to trace but more explicit
	["trace"] = 2,  -- More finely-grained logging than debug (details of calls and execution paths taken)
	["debug"] = 3,  -- Fine-grained logging (primary calls and execution paths taken)
	["info"]  = 4,  -- General information that you'd want to tell the user
	["warn"]  = 5,  -- Something has gone awry or may be missing, but execution can continue or be rewound
	["error"] = 6,  -- Execution must cease immediately, future problems may arise
	["fatal"] = 7,  -- The entire addon must be terminated and stop functioning
	["off"]   = 8,  -- Disable logging (not recommended)
})


local logNames = {}

local outputSources, defaultOutputSource = {}, DEFAULT_CHAT_FRAME
local globalLogLevels, defaultGlobalLogLevel = {}, 4
local currentLogLevels, defaultCurrentLogLevel = {}, 4

local warningCounts = {}
local errorCounts = {}
local maxWarningCounts, defaultMaxWarningCount = {}, 50
local maxErrorCounts, defaultMaxErrorCount = {}, 10

local defaultLogColorizer = TweakerColorizer:new({
	["trace"] = TweakerColorizer:GetColor("lightmagenta"),
	["debug"] = TweakerColorizer:GetColor("lightblue"),
	["info"]  = TweakerColorizer:GetColor("white"),
	["warn"]  = TweakerColorizer:GetColor("lightyellow"),
	["error"] = TweakerColorizer:GetColor("lightred"),
	["fatal"] = TweakerColorizer:GetColor("red"),
})

--====================================================================================================================--

-- The Logger Object ---------------------------------------------------------------------------------------------------

local Logger = {}

------------------------------
-- Properties
------------------------------

function Logger:GetName()
	return logNames[self]
end

function Logger:GetOutputSource()
	return outputSources[self] or defaultOutputSource
end

function Logger:SetOutputSource(source)
	if type(source) ~= "table" or type(source.AddMessage) ~= "function" then
		source = nil
	end
	outputSources[self] = source
end

function Logger:GetGlobalLogLevel()
	return globalLogLevels[self] or defaultGlobalLogLevel
end

function Logger:SetGlobalLogLevel(value)
	value = self:ToLogLevel(value)
	if value then
		globalLogLevels[self] = value
	end
end

function Logger:GetLogLevel()
	return currentLogLevels[self] or defaultCurrentLogLevel
end

function Logger:SetLogLevel(value)
	value = self:ToLogLevel(value)
	if value then
		currentLogLevels[self] = value
	end
end

function Logger:RestoreLogLevel()
	self:SetLogLevel(self:GetGlobalLogLevel())
end

function Logger:GetMaxWarningCount()
	return maxWarningCounts[self] or defaultMaxWarningCount
end

function Logger:SetMaxWarningCount(value)
	maxWarningCounts[self] = value
end

function Logger:GetMaxErrorCount()
	return maxErrorCounts[self] or defaultMaxErrorCount
end

function Logger:SetMaxErrorCount(value)
	maxErrorCounts[self] = value
end

------------------------------
-- Utility Functions
------------------------------

function Logger:ToLogLevel(value)
	if not value then return end
	
	if type(value) == "table" then value = value.LogLevel
	if not value then return end end
	
	value = tonumber(value) or logLevelMatrix[string.lower(value)]
	if not value or value < logLevelMatrix.all or value > logLevelMatrix.off then
		return
	end
	
	return value
end

function Logger:ToLogLevelString(level)
	return tonumber(level) and logLevelMatrix[level]
end

------------------------------
-- User-Defined Functions
------------------------------

Logger.ExcessiveWarningHandler = noop
Logger.ExcessiveErrorHandler = noop
Logger.FatalErrorHandler = noop

------------------------------
-- Logging Functions
------------------------------

function Logger:log(level, logType, logColor, msg, ...)
	if self:GetLogLevel() > level then return end
	
	if (...) then msg = string.format(msg, ...) end
	
	local name = string.prepend(self:GetName(), "-") or ""
	logType = string.prepend(logType, "-") or ""
	--logColor = self.Colorizer:GetColor(logColor) or logColor
	
	util.output(self:GetOutputSource(), self.Colorizer:Apply(logColor, "[Tweaker%s%s] %s", name, logType, msg))
end

function Logger:trace(msg, ...)
	self:log(logLevelMatrix.trace, "Trace", "trace", msg, ...)
end

function Logger:debug(msg, ...)
	self:log(logLevelMatrix.debug, "Debug", "debug", msg, ...)
end

function Logger:info(msg, ...)
	self:log(logLevelMatrix.info, nil, "info", msg, ...)
end

local function logcount(self, counter, max, handler, msg)
	if max == 0 then return end

	counter[self] = counter[self] or {}
	
	local count = (counter[self][msg] or 0) + 1
	counter[self][msg] = count
	
	if count >= max then
		handler(self, count, msg)
	end
end

function Logger:warn(msg, ...)
	if (...) then msg = string.format(msg, ...) end
	self:log(logLevelMatrix.warn, "Warning", "warn", msg)
	
	logcount(self, warningCounts, self:GetMaxWarningCount(), self.ExcessiveWarningHandler, msg)
end

function Logger:error(msg, ...)
	if (...) then msg = string.format(msg, ...) end
	self:log(logLevelMatrix.error, "Error", "error", msg)
	
	logcount(self, errorCounts, self:GetMaxErrorCount(), self.ExcessiveErrorHandler, msg)
	gerror(msg)
end

function Logger:fatal(msg, ...)
	if (...) then msg = string.format(msg, ...) end
	self:log(logLevelMatrix.fatal, "FatalError", "fatal", msg)
	
	self:FatalErrorHandler(msg)
	gerror(msg)
end

function Logger:assert(cond, msg, ...)
	if not cond then self:error(msg, ...) end
	return c
end

------------------------------
-- Constructor
------------------------------

function Logger:new(log)
	log = log or {}
	
	setmetatable(log, self)
	self.__index = self
	
	if log.Name then
		logNames[log] = log.Name
		log.Name = nil
	end
	
	log.Colorizer = defaultLogColorizer:new()
	
	return log
end

--====================================================================================================================--

-- Global --------------------------------------------------------------------------------------------------------------

TweakerLogger = Logger

Tweaker.log = Logger:new()

--====================================================================================================================--
