--====================================================================================================================--
--
-- SLASH COMMAND HANDLER
-- Defines a class for registering and executing slash commands, and a class for defining said slash commands
--
-- Slash commands can be instantiated with TweakerSlashCommand:new()
-- Slash command handlers can be instantiated with TweakerSlashCommandHandler:new()
-- Tweaker's slash command handler is accessible via Tweaker.SlashCommandHandler
--
--====================================================================================================================--

-- Local Data ----------------------------------------------------------------------------------------------------------

local util = Tweaker.util
local string = util.string
local table = util.table

local registeredSlashCommands = {}
local slashCommandHandlerColorizers = {}

local defaultSlashCommandColorizer = TweakerColorizer:new({
	header   = TweakerColorizer:GetColor("lightyellow"),
	command  = TweakerColorizer:GetColor("lightblue"),
	argument = TweakerColorizer:GetColor("orange"),
	invalid  = TweakerColorizer:GetColor("yellow"),
})

--====================================================================================================================--

-- SlashCommand Object -------------------------------------------------------------------------------------------------

local SlashCommand = {
	commands = {},
	help = "",
	args = "",
	call = noop,
}

function SlashCommand:new(command)
	command = command or {}
	command.commands = command.commands or {}
	
	setmetatable(command, self)
	self.__index = self
	
	return command
end

--====================================================================================================================--

-- DefaultHelpSlashCommand ---------------------------------------------------------------------------------------------

local DefaultHelpSlashCommand = SlashCommand:new({
	commands = { "help" },
	help = "Prints the list of supported commands or provides more detailed information on the given command",
	args = "[<any command>]",
	call = function(self, handler, text)
		text = text and string.trim(text) or ""
		
		local c = handler.Colorizer
		
		local header = c:Apply("header", handler.Commands[1])
		
		if text ~= "" then
			local slashcmd = handler:GetCommand(text)
			if slashcmd then
				local aliases = ""
				for i = 2,table.getn(slashcmd.commands) do
					if aliases ~= "" then aliases = aliases..", " end
					aliases = aliases..c:Apply("command", slashcmd.commands[i])
				end
				
				util.outputf("%s %s %s", header, c:Apply("command", slashcmd.commands[1]), slashcmd.args)
				util.output(slashcmd.help)
				util.output(aliases)
			else
				util.output(c:Apply("invalid", "%s is not a valid %s command", text, handler.Name))
			end
		else
			local commands = { "help" }
			local cmds = registeredSlashCommands[handler]
			if cmds then
				for i = 1,table.getn(cmds) do
					table.insert(commands, cmds[i].commands[1])
				end
			end
			table.sort(commands)
			
			util.outputf("%s Commands:", handler.Name)
			for i = 1,table.getn(commands) do
				local cmd = commands[i]
				local slashcmd = handler:GetCommand(cmd)
				util.outputf(" - %s%s", c:Apply("command", cmd), string.prepend(slashcmd.help, ": ") or "")
			end
		end
	end,
})


--====================================================================================================================--

-- The SlashCommandHandler Object --------------------------------------------------------------------------------------

local SlashCommandHandler = {}

------------------------------
-- Properties
------------------------------

function SlashCommandHandler:GetCommand(name)
	-- Default to returning the "help" command
	if not name or name == "" then
		name = "help"
	else
		name = string.lower(name)
	end

	-- Search registered commands first, handlers can define their own, custom help command this way
	local slashCommands = registeredSlashCommands[self]
	if slashCommands then
		for i = 1,table.getn(slashCommands) do
			local slashcmd = slashCommands[i]
			for j = 1,table.getn(slashcmd.commands) do
				if slashcmd.commands[j] == name then
					return slashcmd
				end
			end
		end
	end
	
	-- Nothing yet, see if we're looking for a help command, return the default
	if name == "help" then
		return DefaultHelpSlashCommand
	end
end

------------------------------
-- Functions
------------------------------

function SlashCommandHandler:RegisterCommand(cmd)
	registeredSlashCommands[self] = registeredSlashCommands[self] or {}
	table.insert(registeredSlashCommands[self], cmd)
	return cmd
end

function SlashCommandHandler:CommandRouter(text)
	local command = ""
	while command == "" and text do
		command,text = string.split(" ", text, 2)
	end
	
	if command == "" then
		command = "help"
	else
		command = string.lower(command)
	end
	
	local slashcmd = self:GetCommand(command)
	if slashcmd then
		slashcmd:call(self, text and string.trim(text) or "")
	else
		local cmdName = self.Name and self.Name.." command" or "Command"
		util.output(self.Colorizer:Apply("invalid", cmdName.." not recognized"))
		self:GetCommand("help"):call(self)
	end
end

------------------------------
-- Constructor
------------------------------

function SlashCommandHandler:new(handler)
	handler = handler or {}
	
	setmetatable(handler, self)
	self.__index = self
	
	handler.Colorizer = defaultSlashCommandColorizer:new()
	
	local name = string.upper(handler.Name)
	for i = 1,table.getn(handler.Commands) do
		_G["SLASH_"..name..tostring(i)] = handler.Commands[i]
	end
	SlashCmdList[name] = function(text) handler:CommandRouter(text) end
	
	return handler
end

--====================================================================================================================--

-- Global API ----------------------------------------------------------------------------------------------------------

TweakerSlashCommand = SlashCommand
TweakerSlashCommandHandler = SlashCommandHandler

--====================================================================================================================--

-- Tweaker Setup -------------------------------------------------------------------------------------------------------

local handler = SlashCommandHandler:new({ Name = "Tweaker", Commands = { "/tweaker" } })
Tweaker.SlashCommandHandler = handler

--====================================================================================================================--

-- Tweaker Commands ----------------------------------------------------------------------------------------------------

-- Most commands are defined in their appropriate file.
-- This area is for any that relate to something that needs to be defined prior to Tweaker-Slash

------------------------------
-- Colorize
------------------------------

local colorize_keys = {
	"red", "orange", "yellow",
	"green", "cyan", "blue", "magenta",
	"gray", "white", "black",
}
local colorize_prefixes = { "light", "", "dark" }
Tweaker.SlashCommandHandler:RegisterCommand(SlashCommand:new({
	commands = { "colors" },
	help = "Displays a list of predefined colors that match the given argument, or all colors if no argument is given",
	args = "[<any predefined color>]",
	call = function(self, handler, text)
		text = string.lower(text)
		if text == "" then text = nil end
		
		for i = 1,table.getn(colorize_keys) do
			local k = colorize_keys[i]
			
			for x = 1,table.getn(colorize_prefixes) do
				local pre = colorize_prefixes[x]
				local hex = TweakerColorizer:GetColor(pre..k)
				if hex and ((not text) or string.find(pre..k, text)) then
					util.output("|cff"..hex..pre..k.."|r")
				end
			end
		end
	end,
}))

--====================================================================================================================--