--====================================================================================================================--
--
-- DUMP COMMAND
-- Defines the /dump slash command to dump data in an easy-to-read-and-interpret format for the user
--
-- /dump can be enabled or disabled
-- /tweaker dump is an always-enabled alternative
-- Tweaker.Dump() is the code-accessible alternative, and takes a string
--
--====================================================================================================================--

-- Local Data ----------------------------------------------------------------------------------------------------------

local util = Tweaker.util
local math = util.math
local string = util.string
local table = util.table

local last_dump, current_dump, dump_offset, dump_size, dumping_G, dump_pad, dump_start_time, dumped_tables
local get_global_name, get_global_names
local defaultComparison, tostringComparison, globalNameComparison, comparisonFunctions, comparisonOrder
local get_keys_meta, get_keys, get_sorted_keys
local dumpColorizer
local get_dump_value, get_dump
local dump, old_dump

--====================================================================================================================--

-- Utility Functions ---------------------------------------------------------------------------------------------------

------------------------------
-- Global Name Functions
------------------------------

-- Return's a global's names (in an array)
function get_global_names(value)
	local names = {}
	for k,v in pairs(_G) do
		if v == value and k ~= "arg" then
			table.insert(names, k)
		end
	end
	
	if table.getn(names) == 0 then return end
	
	table.sort(names)
	return names
end

-- Returns a global's names (comma-delimited)
function get_global_name(value)
	local names = get_global_names(value)
	if not names then return end
	
	return string.join(", ", unpack(names))
end

------------------------------
-- Key Comparison Functions
------------------------------

-- Notes:
--  in a key sort, a cannot equal b because all keys must be unique
--  nil cannot be a key -- an error is thrown
--  true and false can be keys, but it'll never compare false to false or true to true
--     false is less than true so if a is false than a is less than b; otherwise b must be false and less than a
--  tables and userdata will be compared by their global name, if applicable. Otherwise, tostring() them

defaultComparison = function(a, b) return a < b end
tostringComparison = function(a, b) return tostring(a) < tostring(b) end
globalNameComparison = function(a, b)
	return (get_global_name(a) or tostring(a)) < (get_global_name(b) or tostring(b))
end

comparisonFunctions = {
	["boolean"] = function(a, b) return a == false end,
	["number"] = defaultComparison,
	["string"] = defaultComparison,
	["function"] = globalNameComparison,
	["table"] = globalNameComparison,
	["thread"] = globalNameComparison,
	["userdata"] = globalNameComparison,
}

comparisonOrder = table.dupekv({ "boolean", "number", "string", "function", "table", "thread", "userdata" })

------------------------------
-- Key Functions
------------------------------

function get_keys_meta(nodupes, t, tcheck, start, trec)
	if not trec or type(trec) ~= "table" then return end
	if tcheck[trec] then return end
	if GetTime() > start + 5 then return end
	
	for k,_ in pairs(trec) do
		if t[k] ~= nil then nodupes[k] = true end
	end
	tcheck[trec] = true
	
	get_keys_meta(nodupes, t, tcheck, start, getmetatable(trec))
	get_keys_meta(nodupes, t, tcheck, start, trec.__index)
end

function get_keys(t)
	local nodupes = {}
	get_keys_meta(nodupes, t, {}, GetTime(), t)
	
	local keys = {}
	for k,_ in pairs(nodupes) do
		table.insert(keys, k)
	end
	
	return keys, table.getn(keys)
end

function get_sorted_keys(t)
	local keys, keyCount = get_keys(t)
	table.sort(keys, function(a,b)
		local ta,tb = type(a),type(b)
		if ta==tb then
			return comparisonFunctions[ta](a, b)
		else
			return comparisonOrder[ta] < comparisonOrder[tb]
		end
	end)
	return keys, keyCount
end

------------------------------
-- COLORS
------------------------------

dumpColorizer = TweakerColorizer:new({
	["nil"]      = TweakerColorizer:GetColor("lightred"),
	["number"]   = TweakerColorizer:GetColor("orange"),
	["string"]   = TweakerColorizer:GetColor("orange"),
	["boolean"]  = TweakerColorizer:GetColor("orange"),
	["table"]    = TweakerColorizer:GetColor("lightblue"),
	["function"] = TweakerColorizer:GetColor("lightgreen"),
	["thread"]   = TweakerColorizer:GetColor("white"),
	["userdata"] = TweakerColorizer:GetColor("white"),
	
	["widget"]   = TweakerColorizer:GetColor("lightyellow"),
	["key"]      = TweakerColorizer:GetColor("lightmagenta"),
})
function dumpColorizer:GetGlobalName(color, value)
	local names = get_global_names(value)
	if not names then return end
	
	local name = ""
	for i = 1,table.getn(names) do
		if string.len(name) > 0 then name = name..", " end
		
		local v = names[i]
		name = name..self:Apply(color, v)
	end
	
	return name
end

------------------------------
-- GET DUMP (TYPE) FUNCTIONS
------------------------------

get_dump_value = {}

get_dump_value["nil"] = function(value)
	return dumpColorizer:Apply("nil", "nil")
end

get_dump_value["number"] = function(value)
	return dumpColorizer:Apply("number", value)
end

get_dump_value["string"] = function(value)
	-- I thought %q was to quote stuff, but it's so it can be read by the interpreter and it's not suitable for
	-- here so I'm quoting and escaping myself
	return dumpColorizer:Apply("string", "%s", string.quote(string.escape(value)))
end

get_dump_value["boolean"] = function(value)
	return dumpColorizer:Apply("boolean", value and "true" or "false")
end

get_dump_value["table"] = function(value)
	if GetTime() > dump_start_time + 5 then return end
	
	local msg = ""
	
	-- Gather General Info --
	
	local is_widget = type(rawget(value, 0)) == "userdata" and type(value.GetObjectType) == "function"
	local t = is_widget and "widget" or "table"
	local keys, keyCount = get_sorted_keys(value)
	local is_first_level = dump_pad == ""
	
	-- Name --
	
	local valueName = dumpColorizer:GetGlobalName(t, value)
	local valueType = tostring(value)
	if is_widget then valueType = value:GetObjectType()..(string.after(valueType, ":", 1) or (" "..valueType)) end
	valueType = dumpColorizer:Apply(t, valueType)
	
	local q1,q2 if not valueName then q1,q2 = "","" elseif is_widget then q1,q2 = "<",">" else q1,q2 = "[","]" end
	msg = msg..(string.space(valueName) or "")..string.quote(valueType, q1, q2)
	
	-- Key Display --
	
	local start_index,stop_index,display_keys
	
	if is_first_level then
		start_index = dump_offset > keyCount and 1 or dump_offset
		stop_index = math.min(keyCount, (start_index + dump_size) - 1)
		display_keys = keyCount > 0 -- always display keys of first level
	else
		if keyCount > dump_size then
			start_index,stop_index = 0,0
			display_keys = false -- too many keys to fit, never display
		else
			start_index,stop_index = 1,keyCount
			display_keys = dumped_tables[value] == nil -- display if it hasn't been dumped already
		end
	end
	--DEFAULT_CHAT_FRAME:AddMessage("str: "..tostring(str))
	--/script print(" ("..a:Apply("nil", "%d Key%s", 0, "s")..")")
	if is_first_level and (start_index > 1 or stop_index < keyCount) then
		msg = msg.." ("..dumpColorizer:Apply("key", "Keys: %d-%d of %d", start_index, stop_index, keyCount)..")"
	else
		msg = msg.." ("..dumpColorizer:Apply("key", "%d Key%s", keyCount, util.plural(keyCount))..")"
	end
	
	dumped_tables[value] = true
	
	-- Display Keys/Values --
	
	if display_keys then
	
		msg = msg..(is_first_level and " =" or "").." {\n"
		
		if stop_index >= start_index then
			dump_pad = dump_pad.."    "
			
			for i = start_index,stop_index do
				local k = keys[i]
				local v = value[k]
				
				local typeK = type(k)
				local keyDisplay
				if typeK == "string" then
					keyDisplay = dumpColorizer:Apply(t, string.escape(k))
				else
					if typeK == "number" or typeK == "boolean" then
						keyDisplay = dumpColorizer:Apply(t, tostring(k))
					else
						keyDisplay = (get_dump_value[typeK](k))
					end
					keyDisplay = dumpColorizer:Apply(t, "[")..keyDisplay..dumpColorizer:Apply(t, "]")
				end
				
				local sep = is_widget and " => " or " = "
				
				msg = msg..dump_pad..keyDisplay..sep..(get_dump_value[type(v)](v)).."\n"
			end
			
			dump_pad = string.sub(dump_pad, 5)
		end
		
		msg = msg..dump_pad.."}"..(is_first_level and "\n" or "")
	end
	
	return msg
end

get_dump_value["function"] = function(value)
	local valueName = string.space(dumpColorizer:GetGlobalName("function", value))
	local valueType = dumpColorizer:Apply("function", tostring(value))
	if not valueName then return valueType end
	return valueName.."("..valueType..")"
end

get_dump_value["thread"] = function(value)
	return "thread"
end

get_dump_value["userdata"] = function(value)
	return "userdata"
end

function get_dump(...)
	local n = select('#', ...)
	local ret = ""
	for i=1,n do
		local value = select(i, ...)
		
		dumping_G = value==g
		last_dump = current_dump
		current_dump = value
		dump_size = tonumber(Tweaker.DumpSize) or 20
		dump_offset = current_dump == last_dump and (dump_offset or 1) + dump_size or 1
		dump_pad = ""
		dump_start_time = GetTime()
		dumped_tables = {}
		
		ret = ret..get_dump_value[type(value)](value).."\n"
	end
	return ret
end

------------------------------
-- MAIN DUMP FUNCTIONS
------------------------------

function dump(...)
	util.output(get_dump(...))
end

------------------------------
-- SLASH COMMAND HANDLING
------------------------------

old_dump = SlashCmdList["DUMP"] or noop
DumpCommandHandler = function(text)
	if Tweaker.UseDump then
		text = string.trim(text)
		local f, err = loadstring("return "..text)
		if not f then
			error(err)
		else
			dump(f())
		end
	else
		old_dump(text)
	end
end
SlashCmdList["DUMP"] = DumpCommandHandler

Tweaker.SlashCommandHandler:RegisterCommand(TweakerSlashCommand:new({
	commands = { "dump", "print", "echo" },
	help = "Interprets and prints the results of the passed values. (Also overrides /dump if Tweaker.UseDump is set)",
	args = "value1[, value2[, value3[, ...[, valueN]]]]",
	call = function(self, handler, text) DumpCommandHandler(text) end,
}))
