--[[

TO DO:

- HookScript support
- HookEvent support
- Add /monitor command to monitor global values

]]

------------------------------------------------------------------------------------------------------------------------
-- LOCALS
------------------------------------------------------------------------------------------------------------------------

-- VARIABLES

local Tweaker = Tweaker
local TweakerEnabled = true
local TweakerSlash = {}
local TweakerSlashHelp = {}
local Tweaks = {}
local TweakQueue = {}
local TweakTimers = {}

local util = Tweaker.util
local string = util.string
local table = util.table

local colorizer = Tweaker.colorizer
local log = Tweaker.log
local slashCommandHandler = Tweaker.SlashCommandHandler


------------------------------------------------------------------------------------------------------------------------
-- TWEAKER
------------------------------------------------------------------------------------------------------------------------

local split, CreateWidgetName
local SetOnUpdate

------------------------------------------------------------
-- UTILITIES
------------------------------------------------------------

function split(...)
	return string.split(" ", ...)
end

function CreateWidgetName(name, parent)
	if name and string.len(name) > 0 then
		return string.gsub(name, "$parent", parent or "")
	end
	return nil
end

------------------------------------------------------------
-- The TweakTimer Metatable
------------------------------------------------------------

local TweakTimerDefaults = {
	__elapsed = 0,
	Interval = 0,
	Function = util.noop
}

local TweakTimer = {
	__index = function(t,k)
		t[k] = TweakTimerDefaults[k]
		return rawget(t,k)
	end
}

------------------------------------------------------------
-- The Tweak Metatable
------------------------------------------------------------

local Tweak
Tweak = {
	Name = "",
	__index = function(t,k)
		return k ~= "__index" and Tweak[k] or nil
	end
}

------------------------------------------------------------
-- Template Inheritance
------------------------------------------------------------

local SkipPropagation = util.noop
local Propagate_Copy = function(s,d,k) d[k] = s[k] end

local Propagate_GetByName = function(s,d,tableName)
	local n=s.Name if n then for k,v in pairs(d[tableName]) do if v.Name==n then return v end end end
end

local PropagateAttributeFunctions = {
	Name = SkipPropagation,
	Virtual = SkipPropagation,
	Templates = SkipPropagation,
	FontStrings = {
		function(s,d) return Propagate_GetByName(s,d,"FontStrings") end,
	},
	Textures = {
		function(s,d) return Propagate_GetByName(s,d,"Textures") end,
	},
	Timers = {
		function(s,d) return Propagate_GetByName(s,d,"Timers") end,
	},
}

local function PropagateAttributes(source, dest, functions, logpadsize)
	logpadsize = logpadsize or 0
	local pad = string.rep(" ", logpadsize)
	
	log:trace(pad.."Propagating %s onto %s", string.quote(source.Name) or "source", string.quote(dest.Name) or "destination")
	for k,v in pairs(source) do
		local s = source
		local d = dest
		local f = functions[k] or Propagate_Copy
		local tf = type(f)
		
		if d[k] == nil and tf == "function" then
			log:trace(pad.."Copying %s", k)
			f(s,d,k)
		elseif k == "Events" then
			log:trace(pad.."Handling events")
			dest[k] = dest[k] or {}
			for sourceEventList,sourceEventFunction in pairs(source[k]) do
				for _,sourceEvent in ipairs({split(sourceEventList)}) do
					sourceEvent = string.upper(sourceEvent)
					local exists = false
					for destEventList, destEventFunction in pairs(dest[k]) do
						for _,destEvent in ipairs({split(destEventList)}) do
							destEvent = string.upper(destEvent)
							if sourceEvent == destEvent then
								exists = true
							end
						end
					end
					if not exists then
						dest[k][sourceEvent] = sourceEventFunction
					end
				end
			end
		elseif tf == "table" then
			log:trace(pad.."Iterating table %s", k)
			d[k] = d[k] or {}
			for tk,tv in pairs(source[k]) do
				s = tv
				d = f[1](s,dest)
 				if not d then
 					d = {}
					table.insert(dest[k], d)
 				end
				PropagateAttributes(s,d,f, logpadsize+4)
			end
		end
	end
end

local function PropagateTweakAttributes(source, dest)
	PropagateAttributes(source, dest, PropagateAttributeFunctions)
end

local function PerformTweakInheritance(tweak)
	if not tweak or not tweak.Templates then return true end
	
	local templates = tweak.Templates
	if type(templates) == "string" then templates = {templates} end
	
	log:debug("Performing inheritance for %s (%s)", tweak.Name, string.join(", ", unpack(templates)))
	
	for index,template in ipairs(templates) do
		local baseTweak = Tweaks[template]
		if baseTweak then
			if PerformTweakInheritance(baseTweak) then
				PropagateTweakAttributes(baseTweak, tweak)
			else
				return false
			end
		else
			log:warn("%s template does not exist yet. Adding %s back into the queue.", template, tweak.Name)
			table.insert(TweakQueue, tweak)
			return false
		end
	end
	
	return true
end

------------------------------------------------------------------------------------------------------------------------
-- DATA
------------------------------------------------------------------------------------------------------------------------

local KeyboardEventAttributes = {
	EnableKeyboard = true,
}
local MouseEventAttributes = {
	EnableMouse = true,
	RegisterForClicks = "AnyUp",
}
local MouseWheelEventAttributes = {
	EnableMouseWheel = true,
}
local MouseDragEventAttributes = {
	RegisterForDrag = "LEFTBUTTON",
	SetMovable = true,
}

local EventAttributeAutomation = {
	ONKEYDOWN = KeyboardEventAttributes,
	ONKEYUP = KeyboardEventAttributes,
	
	ONDRAGSTART = MouseEventAttributes,
	ONDRAGSTOP = MouseEventAttributes,
	ONENTER = MouseEventAttributes,
	ONLEAVE = MouseEventAttributes,
	ONMOUSEDOWN = MouseEventAttributes,
	ONMOUSEUP = MouseEventAttributes,
	
	ONMOUSEWHEEL = MouseWheelEventAttributes,
	
	ONDRAGSTART = MouseDragEventAttributes,
	ONDRAGSTOP = MouseDragEventAttributes,
}

------------------------------------------------------------------------------------------------------------------------
-- METHODS
------------------------------------------------------------------------------------------------------------------------

local function SetAttributes(widget, tweak)

	-- Handle CustomAttributes first in case the user wants to override any of the widget's methods
	if tweak.CustomAttributes then
		for k,v in pairs(tweak.CustomAttributes) do
			log:trace("Custom Attribute %s = %s", k, util.totypestring(v))
			widget[k] = v
		end
	end

	-- Parent
	if tweak.Parent then
		log:trace("Setting parent to %s", tweak.Parent)
		widget:SetParent(tweak.Parent)
	end
	
	-- Points
	if widget.GetPoint then
		local getTweakPoints = function(widget, tweak)
			local parent = _G[tweak.Parent] or (widget.GetParent and widget:GetParent())
			local points = tweak.Points
			if points == nil then return nil, parent end
			
			local ret = {}
			local all_points = nil
			local point_set = nil
			if type(points) ~= "table" then points = util.totable(points) end
			
			local get = function(test, default, ...)
				local n = select('#', ...)
				for i = 1,n do
					local v = select(i, ...)
					if v ~= nil then v = test(v) end
					if v ~= nil then
						return v
					end
				end
				return default
			end
			local offsetTest = function(v) return tonumber(v) end
			local anchorTest = function(v) if type(v) == "string" then v = _G[v] end if v and v.GetObjectType then return v end end
			local anchorPoints = { topleft = true, top = true, topright = true, left = true, center = true,
								   right = true, bottomleft = true, bottom = true, bottomright = true }
			local anchorPointTest = function(v) if anchorPoints[string.lower(v)] then return v end end
			
			local getOffset = function(...) return get(offsetTest, 0, ...) end
			local getAnchor = function(...) return get(anchorTest, nil, ...) end
			local getAnchorPoint = function(...) return get(anchorPointTest, nil, ...) end
			
			for _,point in ipairs(points) do
				local anchorFrom, anchor, anchorTo, xOffset, yOffset = split(point)
				if string.upper(anchorFrom) == "ALL" then
					all_points = _G[anchor] or parent
				else
					local baseAnchor = getAnchorPoint(anchorFrom)
					if baseAnchor then
						ret[baseAnchor] = {
							xOffset = getOffset(xOffset, anchor),
							yOffset = getOffset(yOffset, anchorTo),
							anchor = getAnchor(anchor, parent),
							anchorTo = getAnchorPoint(anchorTo, anchorFrom),
							anchorFrom = baseAnchor,
						}
						point_set = true
					end
				end
			end
			
			if all_points then
				return {{ anchorFrom = "ALL", anchor = all_points }}, parent
			end
			
			if not point_set then
				return nil, parent
			end
			
			return ret, parent
		end
		
		local points, pointParent = getTweakPoints(widget, tweak)
		if not points and widget:GetNumPoints() == 0 then points = {{ anchorFrom = "ALL", anchor = pointParent }} end
		if points then
			widget:ClearAllPoints()
			for _,point in pairs(points) do
				local anchorName = point.anchor and point.anchor:GetName() or "parent"
				if point.anchorFrom == "ALL" then
					log:trace("Setting all points to %s", anchorName)
					widget:SetAllPoints(point.anchor)
				else
					local f,a,t,x,y = point.anchorFrom, point.anchor, point.anchorTo, point.xOffset, point.yOffset
					log:trace("Setting point: %s %s %s %d %d", f, anchorName, t, x, y)
					widget:SetPoint(f,a,t,x,y)
				end
			end
		end
	end
	
	-- Attributes
	local skipAttributeAssignment = util.noop
	local attributeAssignment_Unpack = function(widget, f, k, v)
		log:trace("%s = unpack({%s})", k, string.join(", ", v))
		f(widget, unpack(v))
	end
	local attributeAssignment_Split = function(widget, f, k, v)
		log:trace("%s = split(\"%s\")", k, v)
		f(widget, split(v))
	end
	local attributeAssignment_Assign = function(widget, f, k, v)
		log:trace("%s = %s", k, util.totypestring(v))
		f(widget, v)
	end
	local attributeAssignment_Auto = function(widget, f, k, v)
		local t = type(v)
		if t == "table" then
			attributeAssignment_Unpack(widget, f, k, v)
		elseif t == "string" and string.find(v, " ") then
			attributeAssignment_Split(widget, f, k, v)
		else
			attributeAssignment_Assign(widget, f, k, v)
		end
	end
	
	local attributeAssignmentFunctions = {
		Name = skipAttributeAssignment,
		ObjectType = skipAttributeAssignment,
		
		LogLevel = skipAttributeAssignment,
		
		Virtual = skipAttributeAssignment,
		Templates = skipAttributeAssignment,
		Requirements = skipAttributeAssignment,
		
		Points = skipAttributeAssignment,
		FontStrings = skipAttributeAssignment,
		Textures = skipAttributeAssignment,
		
		Events = skipAttributeAssignment,
		SecureHooks = skipAttributeAssignment,
		Timers = skipAttributeAssignment,
		
		CustomAttributes = skipAttributeAssignment,
		Startup = skipAttributeAssignment,
		
		Parent = attributeAssignment_Assign,
		Text = attributeAssignment_Assign,
		
		Backdrop = function(widget, f, k, v)
			attributeAssignment_Assign(widget, f, k, v)
			
			local keys = { "BackdropColor", "BackdropBorderColor" }
			for i2, k2 in ipairs(keys) do
				local v2 = tweak[k2] or tweak["Set"..k2]
				if v2 then
					local f2 = widget["Set"..k2]
					if f and type(f) == "function" then
						attributeAssignment_Assign(widget, f2, k2, v2)
					end
				end
			end
		end,
	}
	
	for k,v in pairs(tweak) do
		local f = widget["Set"..k] or widget[k]
		if f and type(f) == "function" then
			local f2 = attributeAssignmentFunctions[k] or attributeAssignment_Auto
			f2(widget, f, k, v)
		elseif not attributeAssignmentFunctions[k] then
			log:warn("%s is not a valid attribute for %s", k, tweak.Name)
		end
	end

end

local function ProcessTweak(tweak)
	if not TweakerEnabled then return end
	
	log:SetLogLevel(tweak)
	
	log:debug("Processing %s tweak", tweak.Name)
	
	if not PerformTweakInheritance(tweak) then
		log:RestoreLogLevel()
		return
	end
	
	if tweak.Name and not Tweaks[tweak.Name] then
		Tweaks[tweak.Name] = tweak
	end
	
	if tweak.Virtual then
		log:debug("%s is a virtual tweak, storing for later use", tweak.Name)
		log:RestoreLogLevel()
		return
	end
	
	local widgetParent = tweak.Parent
	local widgetType = tweak.ObjectType or "FRAME"
	local widgetName = CreateWidgetName(tweak.Name, widgetParent)
	
	local widget
	if not _G[widgetName] then
		log:debug("Creating %s %s", string.lower(widgetType), widgetName)
		widget = CreateFrame(widgetType, widgetName, _G[widgetParent], tweak.InheritsFrom)
	else
		widget = _G[widgetName]
	end
	SetAttributes(widget, tweak)
	
	if tweak.Textures then
		for i,data in ipairs(tweak.Textures) do
			log:SetLogLevel(data)
			local name = CreateWidgetName(data.Name, widgetName)
			
			log:debug("Processing texture %s", name)
			
			local child
			if not _G[name] then
				log:debug("Creating texture %s", name)
				child = widget:CreateTexture(name, data.DrawLayer, data.InheritsFrom)
			else
				child = _G[name]
			end
			SetAttributes(child, data)
		end
	end
	
	if tweak.FontStrings then
		for i,data in ipairs(tweak.FontStrings) do
			log:SetLogLevel(data)
			local name = CreateWidgetName(data.Name, widgetName)
			
			log:debug("Processing fontstring %s", name)
			
			local child
			if not _G[name] then
				log:debug("Creating fontstring %s", name)
				child = widget:CreateFontString(name, data.DrawLayer, data.InheritsFrom)
			else
				child = _G[name]
			end
			SetAttributes(child, data)
		end
	end
	
	if tweak.SecureHooks then
		for hooklist,func in pairs(tweak.SecureHooks) do
			for _,hook in pairs({split(hooklist)}) do
				log:trace("Securely hooking %s to %s", hook, tostring(func))
				hooksecurefunc(widget, hook, func)
			end
		end
	end
	
	if tweak.Events then
		local widgetAttributeAutomations = {}
		for eventlist,func in pairs(tweak.Events) do
			for _,event in pairs({split(eventlist)}) do
				event = string.upper(event)
				
				for e,eventAttributes in pairs(EventAttributeAutomation) do
					if event == e then
						for attribute,value in pairs(eventAttributes) do
							widgetAttributeAutomations[attribute] = value
						end
					end
				end
				
				if string.sub(event, 1, 2) == "ON" then
					log:trace("Setting the %s script to %s", event, tostring(func))
					widget:SetScript(event, func)
				else
					widget.__events = widget.__events or {}
					widget.__events[event] = func
					log:trace("Event %s is being registered", event)
					widget:RegisterEvent(event)
				end
				
			end
		end
		
		for attribute,value in pairs(widgetAttributeAutomations) do
			if tweak[attribute] == nil then
				if widget[attribute] then
					log:trace("Auto-assigning attribute %s because tweak was using an event that requires it", attribute)
					widget[attribute](widget, value)
				end
			else
				trace("Unable to auto-assign %s because it's already defined by the tweak", attribute)
			end
		end
		
		if widget.__events then
			if not widget:GetScript("OnEvent") then
				log:trace("Auto-generating OnEvent script in order to handle registered events")
				widget:SetScript("OnEvent", function(self, event, ...)
					self.__events = self.__events or {}
					local f = self.__events[event]
					if type(f) == "function" then
						f(self, event, ...)
					end
				end)
			else
				log:trace("OnEvent script already exists, unable to auto-generate one for registered events")
			end
		end
	end
	
	if tweak.Timers then
		local c = table.getn(tweak.Timers)
		log:trace("Creating %d timer%s for %s", c, util.plural(c), tweak.Name or "tweak")
		for _,timer in ipairs(tweak.Timers) do
			setmetatable(timer, TweakTimer)
			timer.__widget = widget
			timer.__tweak = tweak
			table.insert(TweakTimers, timer)
		end
		SetOnUpdate()
	end
	
	local f = tweak.Startup
	if f then
		log:trace("Calling startup script")
		f(widget)
	end
	
	log:RestoreLogLevel()
	
end


------------------------------------------------------------------------------------------------------------------------
-- MAIN
------------------------------------------------------------------------------------------------------------------------

-- No parent. If parented to UIParent, OnUpdate doesn't fire while the UI is hidden, which could be undesirable
local f = CreateFrame("FRAME", "TweakerFrame", nil)

local req_widget = {}
local req_tweak = {}
local reqs = { widget = req_widget, tweak = req_tweak }
local function TweakMeetsRequirements(tweak)
	if not tweak.Parent and not tweak.Templates and not tweak.Requirements then return true end
	
	for k,v in ipairs(req_widget) do req_widget[k] = nil end
	for k,v in ipairs(req_tweak) do req_tweak[k] = nil end
	
	if tweak.Parent then
		log:trace("%s requires a parent widget named %s", tweak.Name, tweak.Parent)
		table.insert(reqs.widget, tweak.Parent)
	end
	if tweak.Templates then
		log:trace("%s requires the following tweaks for templating: %s", tweak.Name, table.list(tweak.Templates))
		table.insert_multi(reqs.tweak, util.unpack_if_table(tweak.Templates))
	end
	if tweak.Requirements then
		log:trace("%s requires the following widgets: %s", tweak.Name, table.list(tweak.Requirements))
		table.insert_multi(reqs.widget, util.unpack_if_table(tweak.Requirements))
	end
	
	for kind,list in pairs(reqs) do
		local c = (kind == "widget" and _G) or (kind == "tweak" and Tweaks) or _G
		
		for _,v in ipairs(list) do
			if not c[v] then
				log:trace("Required %s %s is still missing, cannot process %s yet", kind, v, tweak.Name)
				return false
			end
		end
	end
	
	log:trace("%s meets its requirements", tweak.Name)
	return true
end

local function RegisterTweak(tweak)
	log:SetLogLevel(tweak)
	setmetatable(tweak, Tweak)
	log:debug("Registering tweak %s", tweak.Name)
	if TweakMeetsRequirements(tweak) then
		ProcessTweak(tweak)
	else
		table.insert(TweakQueue, tweak)
		local n = table.getn(TweakQueue)
		log:debug("%s queued for later processing. %d tweak%s in queue.", tweak.Name, n, util.plural(n))
		SetOnUpdate()
	end
	log:RestoreLogLevel()
end

local function RegisterTweaks(tweaks)
	local tweaks_n = table.getn(tweaks)
	log:debug("Registering %d tweak%s...", tweaks_n, util.plural(tweaks_n) )
	for _,tweak in ipairs(tweaks) do
		RegisterTweak(tweak)
	end
end

local function ProcessTweakQueue()
	if not TweakerEnabled then return end
	
	local n = table.getn(TweakQueue)
	log:trace("Processing %d tweak%s from queue...", n, util.plural(n))
	
	for i = n,1,-1 do
		tweak = TweakQueue[i]
		
		log:SetLogLevel(tweak)
		if TweakMeetsRequirements(tweak) then
			table.remove(TweakQueue, i)
			ProcessTweak(tweak)
			SetOnUpdate()
		end
	end
	log:RestoreLogLevel()
end

--~ local monitors = {}
--~ SLASH_MONITOR1 = "/monitor"
--~ local old_monitor = SlashCmdList["MONITOR"] or function() end
--~ Tweaker.UseMonitor = true
--~ SlashCmdList["MONITOR"] = function(text)
--~ 	if not Tweaker.UseMonitor then
--~ 		old_monitor(text)
--~ 	else
--~ 		text = string.trim(text)
--~ 		if string.len(text) > 0 then
--~ 			table.insert(monitors, { key = text, old_value = nil })
--~ 		end
--~ 	end
--~ end

local function ProcessTimers(elapsed)
	local n = table.getn(TweakTimers)
	for i = n,1,-1 do
		local timer = TweakTimers[i]
		
		timer.__elapsed = timer.__elapsed + elapsed
		if timer.__elapsed > timer.Interval then
			timer.Function(timer.__widget, timer.__elapsed)
			timer.__elapsed = 0
		end
	end
end

local function OnUpdate_Timers(self, elapsed)
	ProcessTimers(elapsed)
end

local function OnUpdate_Tweaks(self, elapsed)
	ProcessTweakQueue()
end

local function OnUpdate(self, elapsed)
	ProcessTimers(elapsed)
	ProcessTweakQueue()
end

function SetOnUpdate()
	local timersExist = table.getn(TweakTimers) > 0
	local tweaksExist = table.getn(TweakQueue) > 0
	
	local func = (timersExist and tweaksExist and OnUpdate) or (timersExist and OnUpdate_Timers) or
				(tweaksExist and OnUpdate_Tweaks) or nil
	local func2 = f:GetScript("OnUpdate")
	
	if func2 ~= func then
		local timerScript,tweakScript = timersExist and "UseTimers" or "", tweaksExist and "UseTweaks" or ""
		log:trace("Setting OnUpdate script for Tweaker to %s%s", timerScript, tweakScript)
		f:SetScript("OnUpdate", func)
	end
end

--~ local searchTime = 0

--~ f:SetScript("OnUpdate", function(self, elapsed)
--~ 	ProcessPanelQueue()
--~ 	searchTime = searchTime + elapsed
--~ 	if searchTime > 1000 then
--~ 		searchTime = 0
--~ 		local keyCount = 0
--~ 		for k,v in pairs(panelqueue) do
--~ 			keyCount = keyCount + 1
--~ 		end
--~ 		if keyCount == 0 then
--~ 			DPrint(2, "Queue is finished.")
--~ 			f:SetScript("OnUpdate", nil)
--~ 		end
--~ 	end
--~ end)

--~ hooksecurefunc("CreateFrame", function(frameType, frameName, parentFrame, inheritsFrame)
--~ 	DPrint(2, "CreateFrame("..frameType..", "..(frameName or "")..") with "..tostring(queueKeyCount).." left in queue")
--~ 	if queueKeyCount > 0 then
--~ 		local frameName = frameName or ""
--~ 		local panels = panelqueue[frameName] or nil
--~ 		if panels then
--~ 			DPrint(2, frameName.." exists.")
--~ 			for _,panelq in ipairs(panels) do
--~ 				CreatePanel(panelq)
--~ 			end
--~ 			panelqueue[frameName] = nil
--~ 			queueKeyCount = queueKeyCount - 1
--~ 		end
--~ 	end
--~ end)

--~ ProcessPanelQueue()

------------------------------------------------------------------------------------------------------------------------
-- GLOBAL API
------------------------------------------------------------------------------------------------------------------------

-- Get/Set Log Levels
Tweaker.SetGlobalLogLevel = function(level)
	log:SetGlobalLogLevel(level)
end
Tweaker.GetGlobalLogLevel = function() local g = log:GetGlobalLogLevel() return g, log:ToLogLevelString(g) end
Tweaker.GetCurrentLogLevel = function() local c = log:GetCurrentLogLevel() return c, log:ToLogLevelString(c) end

-- Logging Functions
Tweaker.trace = function(...) log:trace(...) end
Tweaker.debug = function(...) log:debug(...) end
Tweaker.info = function(...) log:info(...) end
Tweaker.warn = function(...) log:warn(...) end
Tweaker.error = function(...) log:error(...) end
Tweaker.fatal = function(...) log:fatal(...) end
Tweaker.assert = function(...) log:assert(...) end

-- /dump Functions and Settings
Tweaker.Dump = dump
Tweaker.UseDump = true
Tweaker.DumpSize = 20

-- Expose tweaks
Tweaker.Tweaks = Tweaks

-- Registration Functions
Tweaker.RegisterTweak = RegisterTweak
Tweaker.RegisterTweaks = RegisterTweaks
