--====================================================================================================================--
--
-- UTILITY FUNCTIONS
-- Standard utility functions to be used by Tweaker
--
--====================================================================================================================--

-- Utility Declaration -------------------------------------------------------------------------------------------------

local util = {}

--====================================================================================================================--

-- General Utility Functions -------------------------------------------------------------------------------------------

-- For when an empty function is needed, this better than creating a bunch of empty functions
util.noop = function() end

-- Taking a page out of django's book to make it easier to supply plurals
util.plural = function(n,s) s = s or "s" return n == 1 and "" or s end

-- Unpacks the value if it's a table; otherwise, the value is just returned
util.unpack_if_table = function(v) if type(v) == "table" then return unpack(v) end return v end

--====================================================================================================================--

-- Output Utility Functions --------------------------------------------------------------------------------------------

local function output_interpret(...)
	local c,i,arg = DEFAULT_CHAT_FRAME,1,select(1,...)
	if type(arg) == "table" then c = arg; i = i + 1 end
	return c,i
end

local function output_print(c,s, ...) for s in string.gmatch(s, "[^\n]+") do c:AddMessage(s, ...) end end

-- Output to the given widget, or the default one if none given
util.output = function(...) local c,i = output_interpret(...) output_print(c,select(i,...)) end

-- Output using string.format instead of the standard AddMessage parameters
util.outputf = function(...) local c,i = output_interpret(...) s = string.format(select(i, ...)) output_print(c,s) end

--====================================================================================================================--

-- Math Utility Functions ----------------------------------------------------------------------------------------------

util.math = {}
for k,v in pairs(math) do
	util.math[k] = v
end

--====================================================================================================================--

-- String Utility Functions --------------------------------------------------------------------------------------------

util.string = {}
for k,v in pairs(string) do
	util.string[k] = v
end

-- Returns the part of the string that comes after the specified character
util.string.after = function(v,c,i)
	local f,l = string.find(v,c),string.len(c)
	return f and string.sub(v,f+((not i) and l or 0))
end

-- Appends the given characters to the string if the string exists
util.string.append = function(v,c) return v and v..c or nil end

-- Prepends the given characters to the string if the string exists
util.string.prepend = function(v,c) return v and c..v or nil end

-- Quotes a string if a string is given, the "quote" characters can be any string
util.string.quote = function(v,q1,q2) q1 = q1 or '"' q2 = q2 or q1 return v and q1..v..q2 or nil end

-- Appends a space if the string exists
util.string.space = function(v) return v and v.." " or nil end

-- Escapes the string so that it can be displayed the same way it would appear in code
local string_escape_subs = { ["|"] = "||", ["\n"] = "\\n", ["\""] = "\\\"" } -- how to handle \0 ?
util.string.escape = function(s) for k,v in pairs(string_escape_subs) do s = string.gsub(s,k,v) end return s end

--====================================================================================================================--

-- Table Utility Functions ---------------------------------------------------------------------------------------------

util.table = {}
for k,v in pairs(table) do
	util.table[k] = v
end

-- Duplicates a table's values as new table keys with values of the old keys in order to allow for reverse lookups
-- This function modifies the table that's passed as t and returns a reference to it
util.table.dupekv = function(t) for k,v in pairs(t) do t[v] = k end return t end

-- Inserts multiple values into a table, in sequence
util.table.insert_multi = function(t, ...) for i=1,select('#',...) do local v = select(i,...) table.insert(t, v) end end

-- Lists a table's values in a delimited string
util.table.list = function(t,d) return string.join(d or ", ", util.unpack_if_table(t)) end

--====================================================================================================================--

-- Conversion Utility Functions ----------------------------------------------------------------------------------------

-- Converts a value to a table, if it's not already one
util.totable = function(t) return type(t) == "table" and t or {t} end

-- Converts a value to a string, displaying it based on the type of value it is
util.totypestring = function(v) return type(v) == "string" and util.string.quote(v) or tostring(v) end

--====================================================================================================================--

-- Global --------------------------------------------------------------------------------------------------------------

Tweaker.util = util

--====================================================================================================================--